//----------------------------------------------------------------------//
// NyanFi																//
//  Gitr[A															//
//----------------------------------------------------------------------//
#include "UserFunc.h"
#include "UserMdl.h"
#include "Global.h"
#include "FileInfDlg.h"
#include "GitTag.h"
#include "GenInfDlg.h"
#include "InpExDlg.h"
#include "GitGrep.h"
#include "GitView.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TGitViewer *GitViewer = NULL;

//---------------------------------------------------------------------------
__fastcall TGitViewer::TGitViewer(TComponent* Owner)
	: TForm(Owner)
{
	HistoryLimit  = GIT_DEF_HISTLIMIT;
	LastBrListIdx = LastCmListIdx = LastDfListIdx = -1;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FormCreate(TObject *Sender)
{
	BranchScrPanel = new UsrScrollPanel(BranchPanel, BranchListBox, USCRPNL_FLAG_P_WP | USCRPNL_FLAG_L_WP);
	CommitScrPanel = new UsrScrollPanel(CommitPanel, CommitListBox, USCRPNL_FLAG_P_WP | USCRPNL_FLAG_L_WP);
	DiffScrPanel   = new UsrScrollPanel(DiffPanel,   DiffListBox,   USCRPNL_FLAG_P_WP | USCRPNL_FLAG_L_WP);

	MsgHint = new UsrHintWindow(this);
	AssignScaledFont(MsgHint->Font, HintFont, this);

	WarnList   = new TStringList();
	StatusList = new TStringList();

	Staged     = false;
	KeyHandled = false;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FormShow(TObject *Sender)
{
	Caption = yen_to_delimiter(WorkDir) + " - Gitr[A";

	IniFile->LoadPosInfo(this, DialogCenter);

	setup_ToolBar(FindBar);
	setup_ToolBar(DiffBar);

	set_ListBoxItemHi(BranchListBox, ListFont);
	set_UsrScrPanel(BranchScrPanel);
	set_ListBoxItemHi(CommitListBox, ListFont);
	set_UsrScrPanel(CommitScrPanel);
	set_ListBoxItemHi(DiffListBox,	 ListFont);
	set_UsrScrPanel(DiffScrPanel);

	BranchPanel->Width    = IniFile->ReadScaledIntGen(_T("GitViewBranchWidth"), 200, this);
	RL_Panel->Height      = IniFile->ReadScaledIntGen(_T("GitViewDiffHeight"),	120, this);
	FindCommitEdit->Width = IniFile->ReadScaledIntGen(_T("GitViewFindWidth"),	200, this);

	ShowRBranchAction->Checked  = IniFile->ReadBoolGen(_T("GitViewShowRBranch"),true);
	ShowTagAction->Checked      = IniFile->ReadBoolGen(_T("GitViewShowTag"),	true);
	DescTagAction->Checked      = IniFile->ReadBoolGen(_T("GitViewDescTag"),	false);
	ShowBranchesAction->Checked = IniFile->ReadBoolGen(_T("GitViewShowBranches"));
	ShowRemoteAction->Checked   = IniFile->ReadBoolGen(_T("GitViewShowRemote"));
	ShowAuthorAction->Checked   = IniFile->ReadBoolGen(_T("GitViewShowAuthor"));

	CommitSplitter->Color = get_SplitterCol();
	DiffSplitter->Color   = get_SplitterCol();
	BranchListBox->Color  = get_ListBgCol();
	CommitListBox->Color  = get_ListBgCol();
	DiffListBox->Color	  = get_ListBgCol();

	set_EditColor(FindCommitEdit, true);

	::PostMessage(Handle, WM_FORM_SHOWED, 0, 0);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::WmFormShowed(TMessage &msg)
{
	Repaint();

	AutoCrlf   = SameText(Trim(GitExeStr("config --get core.autocrlf")), "true");
	RetArcFile = EmptyStr;

	UpdateCommitList();
	CommitListBox->SetFocus();
}

//---------------------------------------------------------------------------
void __fastcall TGitViewer::FormClose(TObject *Sender, TCloseAction &Action)
{
	BranchListBox->Clear();
	DiffListBox->Clear();
	ClearCommitList();

	FilterName  = EmptyStr;
	HistoryLimit = GIT_DEF_HISTLIMIT;

	IniFile->SavePosInfo(this);
	IniFile->WriteScaledIntGen(_T("GitViewBranchWidth"),	BranchPanel->Width, this);
	IniFile->WriteScaledIntGen(_T("GitViewDiffHeight"),		RL_Panel->Height, this);
	IniFile->WriteScaledIntGen(_T("GitViewFindWidth"),		FindCommitEdit->Width, this);
	IniFile->WriteBoolGen(_T("GitViewShowRBranch"),		ShowRBranchAction);
	IniFile->WriteBoolGen(_T("GitViewShowTag"),			ShowTagAction);
	IniFile->WriteBoolGen(_T("GitViewDescTag"),			DescTagAction);
	IniFile->WriteBoolGen(_T("GitViewShowBranches"),	ShowBranchesAction);
	IniFile->WriteBoolGen(_T("GitViewShowRemote"),		ShowRemoteAction);
	IniFile->WriteBoolGen(_T("GitViewShowAuthor"),		ShowAuthorAction);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FormDestroy(TObject *Sender)
{
	delete BranchScrPanel;
	delete CommitScrPanel;
	delete DiffScrPanel;
	delete MsgHint;
	delete WarnList;
	delete StatusList;
}

//---------------------------------------------------------------------------
void __fastcall TGitViewer::FormKeyDown(TObject *Sender, WORD &Key, TShiftState Shift)
{
	UnicodeString KeyStr = get_KeyStr(Key, Shift);
	if (SameText(KeyStr, "F5"))
		UpdateLogAction->Execute();
	else
		SpecialKeyProc(this, Key, Shift, _T(HELPTOPIC_GIT) _T("#GitViewer"));
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::BranchPanelResize(TObject *Sender)
{
	BranchScrPanel->UpdateKnob();
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitPanelResize(TObject *Sender)
{
	CommitScrPanel->UpdateKnob();
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffPanelResize(TObject *Sender)
{
	DiffScrPanel->UpdateKnob();
}

//---------------------------------------------------------------------------
//gitsʕ擾
//  KvɉĈꗗXV
//---------------------------------------------------------------------------
UnicodeString __fastcall TGitViewer::GitExeStr(UnicodeString prm)
{
	GitBusy = true;
	UnicodeString cmd = get_tkn(prm, ' ');
	if (SameText("archive", cmd)) CommitListBox->Invalidate();
	std::unique_ptr<TStringList> o_lst(new TStringList());
	DWORD exit_code;
	bool res = GitShellExe(prm, WorkDir, o_lst.get(), &exit_code);
	GitBusy = false;

	UnicodeString ret_str;
	if (res) {
		if (exit_code!=0) {
			msgbox_ERR(o_lst->Text);
		}
		else {
			ret_str = o_lst->Text;
			//ԂXV
			if (StartsText("reset HEAD", prm) || contained_wd_i("add|rm", cmd)) {
				GitBusy = true;
				UnicodeString stt_str = get_post_tab(UpdateStatusList());
				GitBusy = false;
				if (!stt_str.IsEmpty()) {
					UnicodeString stt_wk = get_pre_tab(stt_str);
					UnicodeString stt_ix = get_post_tab(stt_str);

					TListBox *c_lp = CommitListBox;
					int cnt = 0;
					for (int i=0; i<c_lp->Count && cnt<2; i++) {
						git_rec *gp = (git_rec *)c_lp->Items->Objects[i];
						if (gp->is_work) {
							gp->msg = get_pre_tab(stt_str);
							gp->diff_inf = EmptyStr;
							cnt++;
						}
						else if (gp->is_index) {
							gp->msg = get_post_tab(stt_str);
							gp->diff_inf = EmptyStr;
							cnt++;
						}
					}
					c_lp->Invalidate();
					UpdateDiffList(true);
				}
			}
			//R~bgXV
			else if (!contained_wd_i("archive|cat-file|config|difftool|gui|rev-parse|show", cmd)) {
				UpdateCommitList();
			}
		}
	}
	return ret_str;
}
//---------------------------------------------------------------------------
//gitsʂ TStringDynArray Ŏ擾
//  G[\/ ꗗ̍XV͍sȂ/ x𕪗擾
//---------------------------------------------------------------------------
TStringDynArray __fastcall TGitViewer::GitExeStrArray(UnicodeString prm)
{
	GitBusy = true;
	std::unique_ptr<TStringList> o_lst(new TStringList());
	DWORD exit_code;
	bool res = GitShellExe(prm, WorkDir, o_lst.get(), &exit_code);
	GitBusy = false;

	TStringDynArray ret_array;
	if (res && exit_code==0) {
		split_GitWarning(o_lst.get(), WarnList);
		for (int i=0; i<o_lst->Count; i++) add_dyn_array(ret_array, o_lst->Strings[i]);
	}
	return ret_array;
}

//---------------------------------------------------------------------------
//gitsʂXgŎ擾 (qg\)
//---------------------------------------------------------------------------
bool __fastcall TGitViewer::GitExeList(UnicodeString prm, TStringList *o_lst, UnicodeString hint)
{
	if (!hint.IsEmpty()) MsgHint->ShowMsgHint(hint, this, col_bgHint);

	GitBusy = true;
	DWORD exit_code;
	bool res = (GitShellExe(prm, WorkDir, o_lst, &exit_code) && exit_code==0);
	GitBusy = false;

	if (!hint.IsEmpty()) MsgHint->ReleaseHandle();
	return res;
}

//---------------------------------------------------------------------------
git_rec * __fastcall TGitViewer::cre_GitRec(UnicodeString msg)
{
	git_rec *gp  = new git_rec;
	gp->is_head  = false;
	gp->is_work  = false;
	gp->is_index = false;
	gp->is_stash = false;
	gp->msg 	 = msg;
	return gp;
}

//---------------------------------------------------------------------------
void __fastcall TGitViewer::ClearCommitList()
{
	TListBox *lp = CommitListBox;
	for (int i=0; i<lp->Count; i++) delete (git_rec*)lp->Items->Objects[i];
	lp->Clear();

	LastBrListIdx = LastCmListIdx = LastDfListIdx = -1;
	MaxGrWidth = lp->ItemHeight;
	MaxAnWidth = 0;

	WarnList->Clear();
	StatusList->Clear();
	Staged = false;
}

//---------------------------------------------------------------------------
//ԃXg̍XV
//  ߂l: Ŝ̏ [TAB ] [LOc[ [TAB] CfbNX
//---------------------------------------------------------------------------
UnicodeString __fastcall TGitViewer::UpdateStatusList()
{
	StatusList->Clear();
	Staged = false;

	std::unique_ptr<TStringList> o_lst(new TStringList());
	if (get_GitStatusList(WorkDir, o_lst.get())>=0) {
		StatusList->Assign(o_lst.get());
		return get_GitStatusStr(o_lst.get(), &Staged);
	}

	return EmptyStr;
}

//---------------------------------------------------------------------------
//R~bgꗗXV
//---------------------------------------------------------------------------
void __fastcall TGitViewer::UpdateCommitList(
	UnicodeString commit_id)	//ΏۃR~bgID	(default = EmptyStr : HEAD)
{
	ClearCommitList();
	CommitScrPanel->HitLines->Clear();

	TListBox *c_lp = CommitListBox;
	MsgHint->ShowMsgHint("t@Co...", c_lp, col_bgHint);
	GitBusy = true;

	std::unique_ptr<TStringList> c_lst(new TStringList());
	std::unique_ptr<TStringList> o_lst(new TStringList());

	//Stash
	UnicodeString prm = "stash list";
	if (GitShellExe(prm, WorkDir, o_lst.get(), NULL, WarnList) && o_lst->Count>0) {
		git_rec *gp;
		for (int i=0; i<o_lst->Count; i++) {
			UnicodeString lbuf = o_lst->Strings[i];
			gp = cre_GitRec();
			gp->graph	 = "*";
			gp->is_stash = true;
			gp->stash	 = split_tkn(lbuf, ": ");
			gp->msg		 = lbuf;
			c_lst->AddObject(EmptyStr, (TObject *)gp);
		}
		//Zp[^
		gp = cre_GitRec();
		gp->graph = "-";
		c_lst->AddObject(EmptyStr, (TObject *)gp);
	}

	//[LOc[/CfbNX
	UnicodeString stt_str = get_post_tab(UpdateStatusList());
	if (!stt_str.IsEmpty()) {
		UnicodeString stt_wk = get_pre_tab(stt_str);
		UnicodeString stt_ix = get_post_tab(stt_str);

		//[LOc[
		git_rec *gp  = cre_GitRec(stt_wk);
		gp->is_work  = true;
		gp->graph	 = "*";

		c_lst->AddObject(EmptyStr, (TObject *)gp);
		//CfbNX
		gp  = cre_GitRec(stt_ix);
		gp->is_index = true;
		gp->graph	 = "*";
		c_lst->AddObject(EmptyStr, (TObject *)gp);
		//Zp[^
		gp  = cre_GitRec();
		gp->graph	= "-";
		c_lst->AddObject(EmptyStr, (TObject *)gp);
	}

	//R~bg
	prm = "log --graph";
	if (ShowBranchesAction->Checked && commit_id.IsEmpty()) prm += " --branches";
	if (HistoryLimit>0) prm.cat_sprintf(_T(" -%u"), HistoryLimit);
	if (!FilterName.IsEmpty()) prm += " --follow";
	prm += " --date=format:\"%Y/%m/%d %H:%M:%S\" --pretty=format:\"\t%H\t%P\t%ad\t%s\t%d\t%an\"";
	if (!commit_id.IsEmpty())  prm.cat_sprintf(_T(" %s"), commit_id.c_str());
	if (!FilterName.IsEmpty()) prm.cat_sprintf(_T(" \"%s\""), FilterName.c_str());

	//Ot / R~bgID / eID /  /  / u` / ^O / Author
	int idx_h = -1;
	UnicodeString last_parent;
	DWORD exit_code;
	o_lst->Clear();
	if (GitShellExe(prm, WorkDir, o_lst.get(), &exit_code, WarnList) && exit_code==0 && o_lst->Count>0) {
		int max_gr_len = 0, max_an_wd = 0;
		for (int i=0; i<o_lst->Count; i++) {
			UnicodeString lbuf = o_lst->Strings[i];
			TStringDynArray ibuf = split_strings_tab(lbuf);
			if (ibuf.Length>0) {
				git_rec *gp = cre_GitRec();
				gp->graph	= ibuf[0];
				max_gr_len	= std::max(max_gr_len, gp->graph.Length());

				UnicodeString br_str;
				if (ibuf.Length==7) {
					gp->hash   = ibuf[1];
					gp->parent = last_parent = ibuf[2];
					gp->msg    = ibuf[4];
					gp->author = ibuf[6];
					max_an_wd = std::max(max_an_wd, c_lp->Canvas->TextWidth(gp->author));

					br_str = get_in_paren(ibuf[5]);
					TStringDynArray b_buf = SplitString(br_str, ",");
					for (int j=0; j<b_buf.Length; j++) {
						UnicodeString ss = Trim(b_buf[j]);
						if (remove_top_text(ss, "tag: ")) {
							if (!gp->tags.IsEmpty()) gp->tags += "\t";
							gp->tags += ss;
						}
						else if (ss.Pos("/")) {
							if (!EndsStr("/HEAD", ss)) gp->branch_r = ss;
						}
						else {
							gp->is_head |= StartsStr("HEAD -> ", ss) || SameStr(ss, "HEAD");
							if (!gp->branch.IsEmpty()) gp->branch += ",";
							gp->branch += ss;
						}
					}

					try {
						gp->f_time = str_to_DateTime(ibuf[3]);
					}
					catch (...) {
						gp->f_time = 0;
					}
				}

				lbuf = gp->hash.SubString(1, 7);
				if (ShowAuthorAction->Checked) lbuf.sprintf(_T(" %s"), gp->author.c_str());
				lbuf.sprintf(_T(" %s %s"), br_str.c_str(), gp->msg.c_str());
				int idx = c_lst->AddObject(lbuf, (TObject *)gp);
				if (gp->is_head) idx_h = idx;
			}
		}

		MaxGrWidth = (max_gr_len + 1) * c_lp->ItemHeight / 3;
		MaxAnWidth = max_an_wd;

		//c̃R~bg
		if (!last_parent.IsEmpty()) {
			prm.sprintf(_T("rev-list --count %s"), last_parent.c_str());
			if (!FilterName.IsEmpty()) prm.cat_sprintf(_T(" -- %s"), add_quot_if_spc(FilterName).c_str());
			o_lst->Clear();
			if (GitShellExe(prm, WorkDir, o_lst.get(), NULL, WarnList) && o_lst->Count>0) {
				int cnt = o_lst->Strings[0].ToIntDef(0);
				if (cnt>0) {
					git_rec *gp = cre_GitRec(
						UnicodeString().sprintf(_T("ccc ȑO %u ̃R~bg܂ ccc"), cnt));
					gp->graph = "#";
					c_lst->AddObject(EmptyStr, (TObject *)gp);
				}
			}
		}
	}

	if (!CommitID.IsEmpty()) {
		for (int i=0; i<c_lst->Count; i++) {
			git_rec *gp = (git_rec *)c_lst->Objects[i];
			if (SameStr(gp->hash, CommitID)) {
				idx_h = i; break;
			}
		}
	}

	if (idx_h==-1) idx_h = 0;

	//Xg{bNXɊ蓖
	c_lp->Items->Assign(c_lst.get());
	SetCommitListIndex(idx_h);
	CommitScrPanel->UpdateKnob();

	GitBusy = false;
	MsgHint->ReleaseHandle();

	UpdateBranchList();

	if (WarnList->Count>0) msgbox_WARN(WarnList->Text);
}

//---------------------------------------------------------------------------
//u`ꗗ̍XV
//---------------------------------------------------------------------------
void __fastcall TGitViewer::UpdateBranchList()
{
	GitBusy = true;

	//[J
	TListBox *b_lp = BranchListBox;
	b_lp->Clear();
	int idx = -1;
	RefHEAD = EmptyStr;
	std::unique_ptr<TStringList> o_lst(new TStringList());
	if (GitShellExe("branch", WorkDir, o_lst.get(), NULL, WarnList)) {
		std::unique_ptr<TStringList> lst(new TStringList());
		lst->AddObject("u`", (TObject *)(NativeInt)(GIT_FLAG_HDRLN|GIT_FLAG_LOCAL));
		for (int i=0; i<o_lst->Count; i++) {
			UnicodeString lbuf = o_lst->Strings[i];
			bool is_hd = remove_top_s(lbuf, '*');
			lst->AddObject(Trim(lbuf), (TObject *)(NativeInt)(GIT_FLAG_LOCAL | (is_hd? GIT_FLAG_HEAD : 0)));
			if (is_hd) {
				idx = lst->Count - 1;
				RefHEAD = lbuf;
			}
		}
		b_lp->Items->Assign(lst.get());
	}
	//[g
	if (ShowRBranchAction->Checked) {
		o_lst->Clear();
		if (GitShellExe("branch -r", WorkDir, o_lst.get(), NULL, WarnList)) {
			std::unique_ptr<TStringList> lst(new TStringList());
			lst->AddObject(EmptyStr, (TObject *)(NativeInt)GIT_FLAG_BLANK);
			lst->AddObject("[g", (TObject *)(NativeInt)(GIT_FLAG_HDRLN|GIT_FLAG_REMOTE));
			for (int i=0; i<o_lst->Count; i++) {
				UnicodeString lbuf = Trim(o_lst->Strings[i]);
				bool is_hd = ContainsStr(lbuf, "/HEAD -> ");
				if (is_hd) lbuf = get_tkn_r(lbuf, "/HEAD -> ");
				if (lst->IndexOf(lbuf)==-1) {
					lst->AddObject(lbuf, (TObject *)(NativeInt)(GIT_FLAG_REMOTE | (is_hd? GIT_FLAG_HEAD : 0)));
				}
			}
			b_lp->Items->AddStrings(lst.get());
		}
	}
	//^O
	if (ShowTagAction->Checked) {
		o_lst->Clear();
		UnicodeString prm = "tag --sort=";
		prm += (DescTagAction->Checked? "-v:refname" : "v:refname"); 
		if (GitShellExe(prm, WorkDir, o_lst.get(), NULL, WarnList)) {
			//R~bgꗗɂ^ÕXg쐬
			std::unique_ptr<TStringList> tag_lst(new TStringList());
			for (int i=0; i<CommitListBox->Count; i++) {
				git_rec *gp = (git_rec *)CommitListBox->Items->Objects[i];
				if (!gp->tags.IsEmpty()) {
					TStringDynArray t_lst = split_strings_tab(gp->tags);
					for (int j=0; j<t_lst.Length; j++) {
					 	if (tag_lst->IndexOf(t_lst[j])==-1) tag_lst->Add(t_lst[j]);
					}
				}
			}
			//^Oڂǉ
			std::unique_ptr<TStringList> lst(new TStringList());
			lst->AddObject(EmptyStr, (TObject *)(NativeInt)GIT_FLAG_BLANK);
			lst->AddObject("^O", (TObject *)(NativeInt)(GIT_FLAG_HDRLN|GIT_FLAG_TAG));
			for (int i=0; i<o_lst->Count; i++) {
				UnicodeString tag = o_lst->Strings[i];
				if (tag.IsEmpty()) continue;
				int flag = GIT_FLAG_TAG;
				if (tag_lst->IndexOf(tag)==-1) flag |= GIT_FLAG_INVAL;
				lst->AddObject(tag, (TObject *)(NativeInt)flag);
			}
			b_lp->Items->AddStrings(lst.get());
		}
	}

	b_lp->ItemIndex = LastBrListIdx = idx;
	BranchScrPanel->UpdateKnob();

	GitBusy = false;
}

//---------------------------------------------------------------------------
void __fastcall TGitViewer::WaitTimerTimer(TObject *Sender)
{
	WaitTimer->Enabled = false;
	UpdateDiffList();
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::UpdateDiffList(
	bool keep_idx)	//ItemIndex ێ	(default = false);
{
	int org_idx = DiffListBox->ItemIndex;
	DiffListBox->Clear();
	CommitID = ParentID = BranchName = TagNames = StashName = EmptyStr;
	MaxDfWidth = 0;

	TListBox *c_lp = CommitListBox;
	git_rec *gp = (git_rec *)c_lp->Items->Objects[c_lp->ItemIndex];
	if (gp->is_stash || gp->is_work || gp->is_index || !gp->hash.IsEmpty()) {
		std::unique_ptr<TStringList> o_lst(new TStringList());
		if (!gp->stash.IsEmpty()) {
			StashName = gp->stash;
			UnicodeString prm = "stash show " + gp->stash;
			GitBusy = true;
			if (GitShellExe(prm, WorkDir, o_lst.get(), NULL, WarnList) && o_lst->Count>0)
				gp->diff_inf = o_lst->Text;
			GitBusy = false;
		}
		else {
			CommitID   = gp->hash;
			ParentID   = gp->parent;
			BranchName = gp->branch;
			TagNames   = gp->tags;

			if (gp->diff_inf.IsEmpty()) {
				UnicodeString prm = "diff --stat-width=120 ";
				UnicodeString parent = get_tkn(ParentID, ' ');
				if (parent.IsEmpty()) parent = GIT_NULL_ID;

				if		(gp->is_work)	;
				else if (gp->is_index)	prm += "--cached";
				else					prm += (parent + " " + CommitID);

				if (!FilterName.IsEmpty()) prm.cat_sprintf(_T(" -- \"%s\""), FilterName.c_str());
				GitBusy = true;
				if (GitShellExe(prm, WorkDir, o_lst.get(), NULL, WarnList) && o_lst->Count>0)
					gp->diff_inf = o_lst->Text;
				GitBusy = false;
			}
			else {
				o_lst->Text = gp->diff_inf;
			}
		}

		//ƃc[ ? t@Cǉ
		if (gp->is_work && StatusList->Text.Pos('?')) {
			o_lst->Add(EmptyStr);
			o_lst->Add("-");
			o_lst->Add("*ǐՂĂȂt@C");
			std::unique_ptr<TStringList> u_lst(new TStringList());
			for (int i=0; i<StatusList->Count; i++) {
				UnicodeString lbuf = StatusList->Strings[i];
				if (lbuf[2]=='?') o_lst->Add(lbuf);
			}
		}

		//Xg{bNXɊ蓖
		TListBox *d_lp = DiffListBox;
		int max_fwd = 0;
		for (int i=0; i<o_lst->Count; i++) {
			UnicodeString lbuf = o_lst->Strings[i];
			if (lbuf.Pos(" | "))
				max_fwd = std::max(max_fwd, d_lp->Canvas->TextWidth(get_tkn(lbuf, " | ") + " "));
		}
		MaxDfWidth = max_fwd;
		d_lp->Items->Assign(o_lst.get());
		d_lp->ItemIndex = keep_idx? std::min(org_idx, d_lp->Count - 1) : 0;
		DiffScrPanel->UpdateKnob();
	}
}

//---------------------------------------------------------------------------
//J[\ʒũu`擾
//---------------------------------------------------------------------------
UnicodeString __fastcall TGitViewer::GetCurBranchName(
	bool lc_only,	//[Ĵ	(default = false)
	bool br_only,	//u`̂	(default = false)
	bool av_only)	//L̂		(default = true)
{
	TListBox *lp = BranchListBox;
	int idx = lp->ItemIndex;
	if (!lp->Focused() || idx==-1) return EmptyStr;

	int flag = (int)lp->Items->Objects[idx];
	if (flag==0) return EmptyStr;
	if (br_only && (flag & GIT_FLAG_TAG))	 return EmptyStr;
	if (lc_only && (flag & GIT_FLAG_REMOTE)) return EmptyStr;
	if (av_only && (flag & GIT_FLAG_INVAL))	 return EmptyStr;
	
	UnicodeString bnam = lp->Items->Strings[idx];
	if (flag & GIT_FLAG_REMOTE) bnam = get_tkn_r(bnam, '/');
	return bnam;
}

//---------------------------------------------------------------------------
//u`ꗗ̕`
//---------------------------------------------------------------------------
void __fastcall TGitViewer::BranchListBoxDrawItem(TWinControl *Control, int Index, TRect &Rect,
	TOwnerDrawState State)
{
	TListBox *lp = (TListBox*)Control;
	int flag = (int)lp->Items->Objects[Index];

	TCanvas *cv = lp->Canvas;
	cv->Brush->Color = State.Contains(odSelected)? (State.Contains(odFocused)? col_selItem : col_oppItem) :
						  (flag & GIT_FLAG_HDRLN)? get_ListHdrBgCol() : lp->Color;
	cv->FillRect(Rect);
	if (flag & GIT_FLAG_BLANK) return;

	int xp = Rect.Left + SCALED_THIS(4);
	int yp = Rect.Top  + SCALED_THIS(2);
	UnicodeString lbuf = lp->Items->Strings[Index];
	//wb_
	if (flag & GIT_FLAG_HDRLN) {
		out_TextEx(cv, xp, yp, lbuf, get_ListHdrFgCol(), get_ListHdrBgCol());
	}
	//
	else {
		//HEAD}[N
		if (flag & GIT_FLAG_HEAD) {
			cv->Font->Color = (flag & GIT_FLAG_REMOTE)? col_GitBraR : col_GitHEAD;
			cv->TextOut(xp, yp, HEAD_Mark);
		}
		xp += cv->TextWidth(HEAD_Mark);
		//u`/^O
		cv->Font->Color = is_SelFgCol(State)? col_fgSelItem : (flag & GIT_FLAG_INVAL)? col_InvItem : get_ListFgCol();
		cv->TextOut(xp, yp, lbuf);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::BranchListBoxClick(TObject *Sender)
{
	TListBox *b_lp = (TListBox*)Sender;
	int idx = b_lp->ItemIndex;	if (idx==-1) return;
	if (idx!=LastBrListIdx) {
		bool found = false;
		if (idx==0) idx = 1;
		bool is_dw = idx>LastBrListIdx;
		for (int i=idx; (is_dw? i<b_lp->Count : i>0) && !found; i+=(is_dw? 1 : -1)) {
			int flag = (int)b_lp->Items->Objects[i];
			if ((flag & (GIT_FLAG_HDRLN|GIT_FLAG_BLANK))==0) {
				idx = i;  found = true;
			}
		}
		if (!found) idx = LastBrListIdx;
		b_lp->ItemIndex = LastBrListIdx = idx;
	}

	TagNames = EmptyStr;

	//R~bg̃J[\ʒuɔf
	if (idx!=-1) {
		int flag = (int)b_lp->Items->Objects[b_lp->ItemIndex];
		UnicodeString lbuf	 = b_lp->Items->Strings[idx];
		UnicodeString b_name = (flag & GIT_FLAG_LOCAL)? lbuf : EmptyStr;
		if (flag & GIT_FLAG_TAG) TagNames = lbuf;

		TListBox *c_lp = CommitListBox;
		if (!b_name.IsEmpty()) {
			for (int i=0; i<c_lp->Count; i++) {
				git_rec *gp = (git_rec *)c_lp->Items->Objects[i];
				if (SameText(b_name, gp->branch)) {
					c_lp->ItemIndex = i;
					c_lp->Invalidate();
					break;
				}
			}
		}
		else if (!TagNames.IsEmpty()) {
			for (int i=0; i<c_lp->Count; i++) {
				git_rec *gp = (git_rec *)c_lp->Items->Objects[i];
				if (contained_wd_i(ReplaceStr(gp->tags, "\t", "|"), TagNames)) {
					c_lp->ItemIndex = i;
					c_lp->Invalidate();
					break;
				}
			}
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::BranchListBoxDblClick(TObject *Sender)
{
	ChckoutAction->Execute();
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::BranchListBoxKeyDown(TObject *Sender, WORD &Key, TShiftState Shift)
{
	TListBox *lp = (TListBox*)Sender;
	UnicodeString KeyStr = get_KeyStr(Key, Shift);
	UnicodeString cmd_F  = Key_to_CmdF(KeyStr);

	if (ExeCmdListBox(lp, cmd_F) || ExeCmdListBox(lp, Key_to_CmdV(KeyStr)))
												BranchListBoxClick(BranchListBox);
	else if (is_ToRightOpe(KeyStr, cmd_F))		CommitListBox->SetFocus();
	else if (SameText(cmd_F, "ReturnList"))		ModalResult = mrCancel;
	else if (SameText(cmd_F, "RenameDlg")) 		RenBranchAction->Execute();
	else if (equal_ENTER(KeyStr)) 				ChckoutAction->Execute();
	else return;

	Key = 0;
}
//---------------------------------------------------------------------------
// R~bg̕`
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitListBoxDrawItem(TWinControl *Control, int Index, TRect &Rect,
	TOwnerDrawState State)
{
	TListBox *lp = (TListBox*)Control;

	UnicodeString kwd = FindCommitEdit->Text;
	bool is_match = !kwd.IsEmpty() && ContainsText(lp->Items->Strings[Index], kwd);

	TCanvas *cv = lp->Canvas;
	cv->Brush->Color = is_match? col_matchItem : lp->Color;
	cv->FillRect(Rect);

	int max_h_wd = 0;
	for (int i=0; i<=0x0f; i++) {
		max_h_wd = std::max(max_h_wd, cv->TextWidth(UnicodeString().sprintf(_T("%x"), i)));
	}

	int s_4 = SCALED_THIS(4);
	int s_8 = SCALED_THIS(8);

	git_rec *gp = (git_rec *)lp->Items->Objects[Index];
	int xp = Rect.Left + s_4;
	int yp = Rect.Top  + SCALED_THIS(2);

	if (SameStr(gp->graph, "#")) {
		cv->Font->Color = col_LineNo;
		out_TextEx(cv, xp, yp, gp->msg);
	}
	if (SameStr(gp->graph, "...")) {
		cv->Font->Color = col_LineNo;
		out_TextEx(cv, xp, yp, "cc");
	}
	//Zp[^
	else if (SameStr(gp->graph, "-")) {
		draw_Separator(cv, Rect);
	}
	//Ot
	else if (!gp->graph.IsEmpty()) {
		git_rec *gp1 = (Index>0)? (git_rec *)lp->Items->Objects[Index - 1] : NULL;
		git_rec *gp2 = (Index<(lp->Count - 1))? (git_rec *)lp->Items->Objects[Index + 1] : NULL;
		UnicodeString s1 = (!gp->is_stash && gp1 && !SameStr(gp1->graph, "-"))? gp1->graph : EmptyStr;
		UnicodeString s2 = (!gp->is_stash && !gp->is_index && gp2)? gp2->graph : EmptyStr;
		TRect rc = Rect; rc.Left = xp;
		draw_GitGraph(gp->graph, s1, s2, cv, rc, gp->is_head, (gp->is_stash || gp->is_work || gp->is_index));
	}

	xp += MaxGrWidth;
	if (gp->is_stash) {
		out_TextEx(cv, xp, yp, gp->stash, col_Folder, col_None, s_8);
		UnicodeString s = gp->msg;
		out_TextEx(cv, xp, yp, split_tkn(s, ": "), col_GitBra, col_None, s_8);
		if (TRegEx::IsMatch(s, "^[0-9a-f]{7}\\s")) out_TextEx(cv, xp, yp, split_tkn(s, ' '), col_GitHash, col_None, s_8);
		out_TextEx(cv, xp, yp, s, get_ListFgCol(), col_None);
	}
	if (gp->is_work || gp->is_index) {
		UnicodeString s1 = "ƃc[ ";
		UnicodeString s2 = "CfbNX ";
		UnicodeString s  = gp->is_work? s1 : s2;
		xp += (std::max(cv->TextWidth(s1), cv->TextWidth(s2)) - cv->TextWidth(s));
		out_TextEx(cv, xp, yp, s, col_Folder, col_None, 0);
		out_TextEx(cv, xp, yp, gp->msg, get_ListFgCol());
	}
	else if (!gp->hash.IsEmpty()) {
		//nbV
		cv->Font->Color = col_GitHash;
		for (int i=1; i<8; i++) {
			UnicodeString c = gp->hash.SubString(i, 1);
			cv->TextOut(xp + (max_h_wd - cv->TextWidth(c))/2, yp, c);
			xp += max_h_wd;
		}
		xp += max_h_wd;

		//Author
		if (ShowAuthorAction->Checked) {
			cv->Font->Color = get_ListFgCol();
			cv->TextOut(xp, yp, gp->author);
			xp += (MaxAnWidth + s_8);
		}

		//
		out_TextEx(cv, xp, yp, get_TimeStampStr(gp->f_time), get_TimeColor(gp->f_time, get_ListFgCol()), 
			col_None, max_h_wd);

		//^O
		if (!gp->tags.IsEmpty()) {
			TStringDynArray t_lst = split_strings_tab(gp->tags);
			for (int i=0; i<t_lst.Length; i++) draw_GitTag(cv, xp, yp, t_lst[i], s_4);
			xp += s_4;
		}

		//u`
		if (!gp->branch.IsEmpty()) {
			TStringDynArray b_buf = SplitString(gp->branch, ",");
			for (int i=0; i<b_buf.Length; i++) {
				UnicodeString ss = b_buf[i];
				if (remove_top_s(ss, "HEAD -> ") || SameStr(ss, "HEAD"))
					out_TextEx(cv, xp, yp, HEAD_Mark, col_GitHEAD, col_None, 0);
				out_TextEx(cv, xp, yp, ss, get_ListBgCol(), col_GitBra, s_4);
				xp += s_4;
			}
		}
		if (ShowRemoteAction->Checked && !gp->branch_r.IsEmpty()) {
			out_TextEx(cv, xp, yp, gp->branch_r, get_ListBgCol(), col_GitBraR, s_8);
		}

		//bZ[W
		out_TextEx(cv, xp, yp, gp->msg, get_ListFgCol());
	}

	draw_ListCursor2(lp, Rect, Index, State);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitListBoxClick(TObject *Sender)
{
	//R~bgł͂ȂڂXLbv
	TListBox *c_lp = CommitListBox;
	if (c_lp->ItemIndex==-1) return;
	if (c_lp->ItemIndex!=LastCmListIdx) {
		int idx = c_lp->ItemIndex;
		bool found = false;
		bool is_dw = idx>LastCmListIdx;
		for (int i=idx; (is_dw? i<c_lp->Count : i>=0) && !found; i+=(is_dw? 1 : -1)) {
			git_rec *gp = (git_rec *)c_lp->Items->Objects[i];
			if (gp->is_stash || gp->is_work || gp->is_index || !gp->hash.IsEmpty()) {
				idx = i;  found = true;
			}
		}
		if (!found) idx = LastCmListIdx;
		c_lp->ItemIndex = LastCmListIdx = idx;
	}

	WaitTimer->Enabled  = false;
	WaitTimer->Interval = REPEAT_WAIT;
	WaitTimer->Enabled  = true;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitListBoxKeyDown(TObject *Sender, WORD &Key, TShiftState Shift)
{
	TListBox *lp = (TListBox*)Sender;
	UnicodeString KeyStr = get_KeyStr(Key, Shift);
	UnicodeString cmd_F  = Key_to_CmdF(KeyStr);

	if (ExeCmdListBox(lp, cmd_F) || ExeCmdListBox(lp, Key_to_CmdV(KeyStr)))
												CommitListBoxClick(NULL);
	else if (is_ToRightOpe(KeyStr, cmd_F))		DiffListBox->SetFocus();
	else if (is_ToLeftOpe(KeyStr, cmd_F))		BranchListBox->SetFocus();
	else if (SameText(cmd_F, "ReturnList"))		ModalResult = mrCancel;
	else if (SameText(cmd_F, "IncSearch"))		FindCommitEdit->SetFocus();
	else if (SameText(cmd_F, "ShowFileInfo") || equal_ENTER(KeyStr))
												CommitInfoAction->Execute();
	else if (SameText(get_CmdStr(cmd_F), "Pack"))
												ArchiveAction->Execute();
	else return;

	Key = 0;
}

//---------------------------------------------------------------------------
//T̕`
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffListBoxDrawItem(TWinControl *Control, int Index, TRect &Rect,
	TOwnerDrawState State)
{
	TListBox *lp = (TListBox*)Control;
	TCanvas  *cv = lp->Canvas;
	cv->Brush->Color = lp->Color;
	cv->FillRect(Rect);

	int xp = Rect.Left + SCALED_THIS(4);
	int yp = Rect.Top  + SCALED_THIS(2);

	UnicodeString lbuf = lp->Items->Strings[Index];
	if (lbuf.Pos(" | ")) {
		UnicodeString s = split_tkn(lbuf, " | ");

		//fBNg
		int p = pos_r('/', s);
		if (p>1) {
			UnicodeString dnam = s.SubString(1, p);
			s.Delete(1, p);
			if (dnam.Pos(" => ")) {
				UnicodeString s1 = split_tkn(dnam, " => ");
				out_TextEx(cv, xp, yp, s1, col_Folder, col_None, 0);
				out_TextEx(cv, xp, yp, "  ", get_ListFgCol(), col_None, 0);
			}
			out_TextEx(cv, xp, yp, dnam, col_Folder, col_None, 0);
		}

		//t@C
		if (s.Pos(" => ")) {
			UnicodeString s1 = split_tkn(s, " => ");
			out_TextEx(cv, xp, yp, s1, get_ExtColor(get_extension(s1), get_ListFgCol()), col_None, 0);
			out_TextEx(cv, xp, yp, "  ", get_ListFgCol());
		}
		out_TextEx(cv, xp, yp, s, get_ExtColor(get_extension(Trim(s)), get_ListFgCol()));

		//r
		TRect rc = Rect; rc.Left = xp = MaxDfWidth;
		RuledLnTextOut("", cv, rc, get_ListFgCol());
		xp = rc.Left;
		//v
		s = split_tkn_spc(lbuf) + " ";
		if (StartsText("Bin", Trim(s))) {
			out_TextEx(cv, xp, yp, s, AdjustColor(get_ListFgCol(), ADJCOL_FGLIST));
			lbuf = ReplaceStr(lbuf, "->", "");
			out_TextEx(cv, xp, yp, lbuf, get_ListFgCol());
		}
		else {
			//ύX
			UnicodeString tmp = ReplaceStr(s, " ", "0");
			xp += (cv->TextWidth(tmp) - cv->TextWidth(s));
			out_TextEx(cv, xp, yp, s, get_ListFgCol());
			//Ot
			int g_w = rc.Height() / 3;
			TRect g_rc = rc;
			g_rc.Left = xp; g_rc.SetWidth(g_w); g_rc.SetHeight(g_w); g_rc.Offset(0, g_w);
			for (int i=1; i<=lbuf.Length(); i++) {
				WideChar c = lbuf[i];
				cv->Brush->Color = (c=='-')? col_GitDel : (c=='+')? col_GitIns : col_None;
				if (cv->Brush->Color!=col_None) {
					cv->FillRect(g_rc);
					g_rc.Offset(g_w, 0);
				}
			}
		}
	}
	//Zp[^
	else if (SameStr(lbuf, "-")) {
		draw_Separator(cv, Rect);
	}
	//ǐՂĂȂ`
	else if (lbuf.Pos('*')) {
		out_TextEx(cv, xp, yp, ReplaceStr(lbuf, "*", "?"), col_Headline);
	}
	//? t@C
	else if (SameStr(lbuf.SubString(2, 1), "?")) {
		lbuf.Delete(1, 2);
		out_TextEx(cv, xp, yp, lbuf, get_ExtColor(get_extension(Trim(lbuf)), get_ListFgCol()));
	}
	//v
	else {
		TStringDynArray t_buf = get_csv_array(lbuf, 3);
		if (t_buf.Length==3) {
			for (int i=0; i<3; i++) {
				UnicodeString s = t_buf[i];
				if (i>0) out_TextEx(cv, xp, yp, ",", get_ListFgCol());
				out_TextEx(cv, xp, yp, s,
					remove_end_text(s, "(-)")? col_GitDel : remove_end_text(s, "(+)")? col_GitIns : get_ListFgCol(),
					col_None, 0);
			}
			draw_separateLine(cv, Rect);
		}
	}

	cv->Brush->Color = lp->Color;
	draw_ListCursor2(lp, Rect, Index, State);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffListBoxEnter(TObject *Sender)
{
	DiffListBoxClick(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffListBoxClick(TObject *Sender)
{
	TListBox *d_lp = DiffListBox;
	if (d_lp->ItemIndex==-1) return;
	if (d_lp->ItemIndex!=LastDfListIdx) {
		int idx = d_lp->ItemIndex;
		bool found = false;
		bool is_dw = idx>LastDfListIdx;
		for (int i=idx; (is_dw? i<d_lp->Count : i>=0) && !found; i+=(is_dw? 1 : -1)) {
			UnicodeString lbuf = d_lp->Items->Strings[i];
			if (lbuf.Pos(" | ") || lbuf.Pos("? ")) {
				idx = i;  found = true;
			}
		}
		if (!found) idx = LastDfListIdx;
		d_lp->ItemIndex = LastDfListIdx = idx;
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffListBoxKeyDown(TObject *Sender, WORD &Key, TShiftState Shift)
{
	TListBox *lp = (TListBox*)Sender;
	UnicodeString KeyStr = get_KeyStr(Key, Shift);
	UnicodeString cmd_F  = Key_to_CmdF(KeyStr);

	if (ExeCmdListBox(lp, cmd_F) || ExeCmdListBox(lp, Key_to_CmdV(KeyStr)))
												DiffListBoxClick(NULL);
	else if (SameText(cmd_F, "ReturnList")) 	ModalResult = mrCancel;
	else if (is_ToLeftOpe(KeyStr, cmd_F))		BranchListBox->SetFocus();
	else if (is_ToRightOpe(KeyStr, cmd_F))		CommitListBox->SetFocus();
	else if (SameText(cmd_F, "FileEdit"))		EditFileAction->Execute();
	else if (SameText(cmd_F, "TextViewer"))		ViewFileAction->Execute();
	else if (SameText(cmd_F, "ShowFileInfo") || equal_ENTER(KeyStr))
												DiffDetailAction->Execute();
	else return;

	Key = 0;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::GitListBoxKeyPress(TObject *Sender, System::WideChar &Key)
{
	Key = 0;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::GitListBoxMouseDown(TObject *Sender, TMouseButton Button,
	TShiftState Shift, int X, int Y)
{
	((TListBox *)Sender)->SetFocus();
}

//---------------------------------------------------------------------------
//`FbNAEg
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CheckoutActionExecute(TObject *Sender)
{
	UnicodeString bnam = GetCurBranchName();
	if (!bnam.IsEmpty()) {
		CommitID = EmptyStr;
		GitExeStr("checkout " + bnam);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CheckoutActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	int flag = GetCurBranchFlag();
	ap->Visible = ((flag & GIT_FLAG_HEAD)==0 || (flag & GIT_FLAG_LOCAL)==0);
	ap->Enabled = ap->Visible && ((flag & GIT_FLAG_INVAL)==0);
}

//---------------------------------------------------------------------------
//u`̍쐬
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CreBranchActionExecute(TObject *Sender)
{
	UnicodeString bnam = InputBox("u`̍쐬", "u`", EmptyStr);
	if (!bnam.IsEmpty()) GitExeStr("branch " + bnam);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CreBranchActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = CurIsHEAD();
}

//---------------------------------------------------------------------------
//u`̍폜
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DelBranchActionExecute(TObject *Sender)
{
	UnicodeString bnam = GetCurBranchName(true, true);
	if (!bnam.IsEmpty()) {
		UnicodeString msg;
		msg.sprintf(_T("u`[%s]폜Ă悢ł?"), bnam.c_str());
		Screen->MessageFont->Assign(Font);
		TForm *MsgDlg = CreateMessageDialog(msg, mtConfirmation,
							TMsgDlgButtons() << mbYes << mbNo << mbCancel, mbNo);
		TCheckBox *cp = new TCheckBox(MsgDlg);
		MsgDlg->ClientHeight = MsgDlg->ClientHeight + cp->Height + 12;
		cp->Caption = "Iɍ폜(&D)";
		cp->Parent	= MsgDlg;
		cp->Left	= 20;
		cp->Top		= MsgDlg->ClientHeight - cp->Height - 12;
		cp->Width	= MsgDlg->ClientWidth - cp->Left;

		if (MsgDlg->ShowModal()==mrYes) {
			UnicodeString prm;
			prm.sprintf(_T("branch -%s %s"), (cp->Checked? _T("D") : _T("d")), bnam.c_str());
			GitExeStr(prm);
		}

		delete MsgDlg;
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DelBranchActionUpdate(TObject *Sender)
{
	int flag = GetCurBranchFlag();
	((TAction *)Sender)->Visible = !(flag & GIT_FLAG_HEAD) && (flag & GIT_FLAG_LOCAL);
}

//---------------------------------------------------------------------------
//u`̖OύX
//---------------------------------------------------------------------------
void __fastcall TGitViewer::RenBranchActionExecute(TObject *Sender)
{
	UnicodeString bnam = GetCurBranchName(true, true, false);
	if (!bnam.IsEmpty()) {
		UnicodeString new_nam = InputBox("u`̖OύX", "u`", bnam);
		if (!new_nam.IsEmpty()) {
			UnicodeString prm;
			prm.sprintf(_T("branch -m %s %s"), bnam.c_str(), new_nam.c_str());
			GitExeStr(prm);
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::RenBranchActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = !GetCurBranchName(true, true, false).IsEmpty();
}

//---------------------------------------------------------------------------
//^Ot
//---------------------------------------------------------------------------
void __fastcall TGitViewer::SetTagActionExecute(TObject *Sender)
{
	if (!SetGitTagDlg) SetGitTagDlg = new TSetGitTagDlg(this);	//ɓI쐬
	SetGitTagDlg->CommitID	  = BranchListBox->Focused()? GetCurBranchName(true, true) : CommitID;
	SetGitTagDlg->EditMsgFile = IncludeTrailingPathDelimiter(WorkDir) + ".git\\TAG_EDITMSG";
	if (SetGitTagDlg->ShowModal()==mrOk) GitExeStr(SetGitTagDlg->GitParam);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::SetTagActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible =
		BranchListBox->Focused()? !GetCurBranchName(true, true).IsEmpty() :
		CommitListBox->Focused()? !CommitID.IsEmpty() : false;
}

//---------------------------------------------------------------------------
//^O̍폜
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DelTagActionExecute(TObject *Sender)
{
	UnicodeString tag = get_pre_tab(TagNames);
	if (!tag.IsEmpty()) {
		UnicodeString msg;
		msg.sprintf(_T("^O [%s] 폜Ă悢ł?"), tag.c_str());
		if (msgbox_Sure(msg)) GitExeStr("tag -d " + tag);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DelTagActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = !get_pre_tab(TagNames).IsEmpty();
}

//---------------------------------------------------------------------------
//}[W
//---------------------------------------------------------------------------
void __fastcall TGitViewer::MergeActionExecute(TObject *Sender)
{
	UnicodeString bnam = GetCurBranchName(true, true);
	if (!CurIsHEAD() && !bnam.IsEmpty()) {
		UnicodeString msg;
		msg.sprintf(_T("[%s]  [%s] Ƀ}[W܂?"), bnam.c_str(), RefHEAD.c_str());
		if (msgbox_Sure(msg)) GitExeStr("merge " + bnam);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::MergeActionUpdate(TObject *Sender)
{
	int flag = GetCurBranchFlag();
	((TAction *)Sender)->Visible = !(flag & GIT_FLAG_HEAD) && (flag & GIT_FLAG_LOCAL);
}
//---------------------------------------------------------------------------
//`F[sbN
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CherryPickActionExecute(TObject *Sender)
{
	git_rec *gp = GetCurCommitPtr();
	if (!SameText(gp->branch, RefHEAD)) {
		if (msgbox_Sure("̃R~bg`F[sbN܂?")) GitExeStr("cherry-pick " + gp->hash);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CherryPickActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = !CommitID.IsEmpty();
}

//---------------------------------------------------------------------------
//u`Rs[
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CopyBranchNameActionExecute(TObject *Sender)
{
	TStringDynArray b_buf = SplitString(BranchName, ",");
	UnicodeString bnam;
	for (int i=0; i<b_buf.Length; i++) {
		UnicodeString ss = b_buf[i];
		remove_top_s(ss, "HEAD -> ");
		remove_top_s(ss, "HEAD");
		if (!ss.IsEmpty()) {
			if (!bnam.IsEmpty()) bnam += ",";
			bnam += ss;
		}
	}
	copy_to_Clipboard(bnam);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CopyBranchNameActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = !BranchName.IsEmpty();
}
//---------------------------------------------------------------------------
//R~bgIDRs[
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CopyCommitIDActionExecute(TObject *Sender)
{
	copy_to_Clipboard(CommitID);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CopyCommitIDActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = !CommitID.IsEmpty();
}

//---------------------------------------------------------------------------
//ÕR~bgɖ߂
//---------------------------------------------------------------------------
void __fastcall TGitViewer::BrPopupMenuPopup(TObject *Sender)
{
	ResetItem->Visible = CurIsHEAD();
	reduction_MenuLine(((TPopupMenu*)Sender)->Items);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ResetItemClick(TObject *Sender)
{
	int tag = ((TMenuItem *)Sender)->Tag;
	UnicodeString prm = (tag==2)? "Hard" : (tag==0)? "Soft" : "Mixed";
	UnicodeString msg;
	msg.sprintf(_T("[%s] 𒼑ÕR~bgɖ߂܂(%s)?"), RefHEAD.c_str(), prm.c_str());
	if (msgbox_Sure(msg)) GitExeStr("reset --" + prm.LowerCase() + " HEAD^");
}

//---------------------------------------------------------------------------
//A[JCu쐬
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ArchiveActionExecute(TObject *Sender)
{
	if (!CommitID.IsEmpty()) {
		UserModule->PrepareSaveDlg(_T("A[JCu̍쐬"), _T("A[JCu (*.zip)|*.zip"), _T("archive.zip"), WorkDir);
		UnicodeString arc_name = UserModule->SaveDlgExecute();
		if (!arc_name.IsEmpty()) {
			UnicodeString prm;
			prm.sprintf(_T("archive --format=zip %s --output=\"%s\""), CommitID.c_str(), yen_to_slash(arc_name).c_str());
			GitExeList(prm, NULL, "A[JCu쐬...");
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ArchiveActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = !CommitID.IsEmpty();
}
//---------------------------------------------------------------------------
//ꎞA[JCuƂĊJ
//---------------------------------------------------------------------------
void __fastcall TGitViewer::OpenTmpArcActionExecute(TObject *Sender)
{
	if (!CommitID.IsEmpty()) {
		UnicodeString tmp_name = TempPathA + CommitID.SubString(1, 8) + "_" + get_dir_name(WorkDir) + ".zip";
		UnicodeString prm;
		prm.sprintf(_T("archive --format=zip %s --output=\"%s\""), CommitID.c_str(), yen_to_slash(tmp_name).c_str());
		if (GitExeList(prm, NULL, "A[JCu쐬...") && file_exists(tmp_name)) {
			RetArcFile	= tmp_name;
			ModalResult = mrOk;
		}
	}
}

//---------------------------------------------------------------------------
//̃R~bg܂ł̗\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::LogThisCommitActionExecute(TObject *Sender)
{
	UpdateCommitList(CommitID);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::LogThisCommitActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Visible = !CommitID.IsEmpty();
}

//---------------------------------------------------------------------------
//eR~bgblame
//---------------------------------------------------------------------------
void __fastcall TGitViewer::BlameActionExecute(TObject *Sender)
{
	UnicodeString fnam = GetDiffFileName();
	if (!fnam.IsEmpty()) {
		fnam = get_GitDiffFile2(fnam);

		UnicodeString prm;
		prm.sprintf(_T("gui blame %s %s"), get_tkn(ParentID, ' ').c_str(), fnam.c_str());
		GitExeStr(prm);
	}
}
//---------------------------------------------------------------------------
//̃t@C̃R~bg\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::LogThisFileActionExecute(TObject *Sender)
{
	FilterName = GetDiffFileName();
	UpdateCommitList();
}

//---------------------------------------------------------------------------
//R~bg\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitInfoActionExecute(TObject *Sender)
{
	std::unique_ptr<TStringList> i_lst(new TStringList());
	WarnList->Clear();

	//{
	UnicodeString prm =
		"show --no-patch --date=format:\"%Y/%m/%d %H:%M:%S\" --pretty=format:\""
		"R~bg=%H\n"
		"c[=%T\n"
		"e=%P\n\n"
		"=%an\nEmail=%ae\nғt=%ad\n"
		"R~b^=%cn\nR~b^Email=%ce\nR~b^t=%cd\n\n"
		"=%s\n\"";
	prm.cat_sprintf(_T(" %s"), CommitID.c_str());
	TStringDynArray r_buf = GitExeStrArray(prm);
	if (r_buf.Length>0) {
		for (int i=0; i<r_buf.Length; i++) {
			UnicodeString nstr = get_tkn(r_buf[i], "=");
			if (nstr.IsEmpty()) {
				i_lst->Add(EmptyStr);
			}
			else {
				UnicodeString vstr = get_tkn_r(r_buf[i], '=');
				if (vstr.IsEmpty()) continue;
				int flag = (i<3)? LBFLG_GIT_HASH : EndsStr("t", nstr)? LBFLG_TIME_FIF : 0;
				if (SameText(nstr, "e")) {
					for (int j=0; !vstr.IsEmpty(); j++) {
						add_PropLine((j==0)? nstr : UnicodeString(" "), split_tkn_spc(vstr), i_lst.get(), flag);
					}
				}
				else {
					add_PropLine(nstr, vstr, i_lst.get(), flag);
				}
			}
		}
	}

	//bZ[W
	prm.sprintf(_T("show --no-patch --pretty=medium %s"), CommitID.c_str());
	r_buf = GitExeStrArray(prm);
	if (r_buf.Length>0) {
		int cnt = 0;
		for (int i=0; i<r_buf.Length; i++) {
			UnicodeString lbuf = TrimLeft(r_buf[i]);
			if (cnt==0 && lbuf.IsEmpty()) {
				cnt++;
			}
			else if (cnt>0) {
				cnt++;
				if (cnt==2 || (cnt==3 && lbuf.IsEmpty())) continue;
				add_PropLine(" ", lbuf, i_lst.get());
			}
		}
	}

	//
	prm.sprintf(_T("notes show %s"), CommitID.c_str());
	r_buf = GitExeStrArray(prm);
	if (r_buf.Length>0) {
		for (int i=0; i<r_buf.Length; i++) {
			UnicodeString lbuf = r_buf[i];
			if (r_buf.Length==1) lbuf = Trim(lbuf);
			add_PropLine((i==0)? "" : " ", lbuf, i_lst.get());
		}
	}

	i_lst->Add(EmptyStr);

	//u`
	prm.sprintf(_T("branch --contains %s"), CommitID.c_str());
	r_buf = GitExeStrArray(prm);
	if (r_buf.Length>0) {
		for (int i=0; i<r_buf.Length; i++) {
			UnicodeString lbuf = r_buf[i];
			if (r_buf.Length==1) lbuf = Trim(lbuf);
			add_PropLine((i==0)? "u`" : " ", lbuf, i_lst.get());
		}
	}

	//^O
	if (!TagNames.IsEmpty()) {
		TStringDynArray t_lst = split_strings_tab(TagNames);
		for (int i=0; i<t_lst.Length; i++) {
			add_PropLine("^O", t_lst[i], i_lst.get(), LBFLG_GIT_TAG);
			//
			prm = "show --no-patch --oneline " + t_lst[i];
			r_buf = GitExeStrArray(prm);
			if (r_buf.Length>0) {
				int cnt = 0;
				for (int i=0; i<r_buf.Length; i++) {
					UnicodeString lbuf = r_buf[i];
					if (cnt==0 && lbuf.IsEmpty()) {
						cnt++;
					}
					else if (cnt>0) {
						if (lbuf.IsEmpty()) break;
						if (StartsStr(get_tkn(lbuf, ' '), CommitID)) continue;
						add_PropLine(" ", lbuf, i_lst.get());
						cnt++;
					}
				}
			}
		}
	}
	//ŐV^O
	else {
		prm.sprintf(_T("describe --tags --abbrev=0 %s"), CommitID.c_str());
		r_buf = GitExeStrArray(prm);
		if (r_buf.Length>0) {
			add_PropLine("ŐV^O", r_buf[0], i_lst.get(), LBFLG_GIT_TAG);
		}
	}

	//ΎQ
	prm.sprintf(_T("name-rev --name-only %s"), CommitID.c_str());
	r_buf = GitExeStrArray(prm);
	if (r_buf.Length>0) {
		add_PropLine("ΎQ", r_buf[0], i_lst.get(),
			StartsStr("tags/", r_buf[0])? LBFLG_GIT_TAG : 0);
	}

	//x
	if (WarnList->Count>0) {
		i_lst->Add(EmptyStr);
		i_lst->AddStrings(WarnList);
		WarnList->Clear();
	}

	if (i_lst->Count>0) {
		FileInfoDlg->ItemList->Assign(i_lst.get());
		FileInfoDlg->isGitInfo = true;
		FileInfoDlg->inhNxtPre = true;
		FileInfoDlg->ShowModal();
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitInfoActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	ap->Visible = !CommitID.IsEmpty();
	ap->Enabled = ap->Visible;
}

//---------------------------------------------------------------------------
//̃t@C̃nbVRs[
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CopyFileHashActionExecute(TObject *Sender)
{
	UnicodeString prm;
	prm.sprintf(_T("rev-parse %s:%s"), CommitID.c_str(), GetDiffFileName().c_str());
	copy_to_Clipboard(GitExeStr(prm));
}
//---------------------------------------------------------------------------
//̃t@C̃pXRs[
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CopyPathActionExecute(TObject *Sender)
{
	copy_to_Clipboard(get_GitDiffFile2(GetDiffTextName()));
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CopyFileInfActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	ap->Visible = !CommitID.IsEmpty() && !GetDiffFileName().IsEmpty();
	ap->Enabled = ap->Visible;
}

//---------------------------------------------------------------------------
//̃t@C̍ڍׂ\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffDetailActionExecute(TObject *Sender)
{
	git_rec *gp = GetCurCommitPtr();		if (!gp) return;
	UnicodeString fnam = GetDiffFileName();	if (fnam.IsEmpty()) return;

	UnicodeString fnam2;
	UnicodeString fnam1 = get_GitDiffFiles(fnam, fnam2);
	UnicodeString prm = "diff";

	if (gp->is_stash) {
		prm.cat_sprintf(_T(" %s %s"), gp->stash.c_str(), fnam2.c_str());
	}
	else if (gp->is_work) {
		prm.cat_sprintf(_T(" -- %s"), fnam1.c_str());
	}
	else if (gp->is_index) {
		prm.cat_sprintf(_T(" --cached -- %s"), fnam1.c_str());
	}
	else {
		UnicodeString parent = get_tkn(ParentID, ' ');
		if (parent.IsEmpty()) parent = GIT_NULL_ID;
		if (SameText(fnam1, fnam2)) {
			prm.cat_sprintf(_T(" %s %s"), parent.c_str(), CommitID.c_str());
			prm.cat_sprintf(_T(" -- %s"), add_quot_if_spc(fnam1).c_str());
		}
		else {
			prm.cat_sprintf(_T(" %s:%s %s:%s"),
				parent.c_str(), fnam1.c_str(), CommitID.c_str(), fnam2.c_str());
		}
	}

	GitBusy = true;
	std::unique_ptr<TStringList> o_lst(new TStringList());
	bool ok = (GitShellExe(prm, WorkDir, o_lst.get()) && o_lst->Count>0);
	GitBusy = false;
	if (ok) {
		UnicodeString tit = "$ git " + sha1_to_short(prm);
		o_lst->Insert(0, tit);
		GeneralInfoDlg->Caption = tit.sprintf(_T("ڍ - %s"), fnam2.c_str());
		GeneralInfoDlg->fromGitView = true;
		if (gp->is_work || gp->is_index)
			GeneralInfoDlg->FileName = IncludeTrailingPathDelimiter(WorkDir) + slash_to_yen(fnam1);
		GeneralInfoDlg->GenInfoList->Assign(o_lst.get());
		GeneralInfoDlg->ShowModal();
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffDetailActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Enabled = !GetDiffFileName().IsEmpty();
}
//---------------------------------------------------------------------------
//̃t@C̓e\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ViewFileActionExecute(TObject *Sender)
{
	git_rec *gp = GetCurCommitPtr();	if (!gp) return;

	UnicodeString fnam = GetDiffTextName();
	if (!fnam.IsEmpty()) {
		fnam = get_GitDiffFile2(fnam);

		bool ok;
		std::unique_ptr<TStringList> o_lst(new TStringList());
		if (gp->is_work) {
			fnam = IncludeTrailingPathDelimiter(WorkDir) + slash_to_yen(fnam);
			ok = load_text_ex(fnam, o_lst.get())!=0;
		}
		else {
			UnicodeString prm = "cat-file -p";
			if (gp->is_stash)
				prm.cat_sprintf(_T(" %s:./%s"), gp->stash.c_str(), fnam.c_str());
			else if (gp->is_index)
				prm.cat_sprintf(_T(" :%s"), fnam.c_str());
			else
				prm.cat_sprintf(_T(" %s:%s"), CommitID.c_str(), fnam.c_str());

			ok = GitExeList(prm, o_lst.get(), "t@Co...");
		}

		if (ok) {
			GeneralInfoDlg->Caption     = UnicodeString().sprintf(_T("t@Ce - %s"), fnam.c_str());
			GeneralInfoDlg->FileName    = fnam;
			GeneralInfoDlg->isNonFile   = !gp->is_work;
			GeneralInfoDlg->fromGitView = true;
			GeneralInfoDlg->fromPopWnd  = true;
			GeneralInfoDlg->GenInfoList->Assign(o_lst.get());
			GeneralInfoDlg->ShowModal();
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ViewFileActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Enabled = !GetDiffTextName().IsEmpty();
}

//---------------------------------------------------------------------------
//[gu`\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ShowRBranchActionExecute(TObject *Sender)
{
	ShowRBranchAction->Checked = !ShowRBranchAction->Checked;
	UpdateBranchList();
}
//---------------------------------------------------------------------------
//^O\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ShowTagActionExecute(TObject *Sender)
{
	ShowTagAction->Checked = !ShowTagAction->Checked;
	UpdateBranchList();
}
//---------------------------------------------------------------------------
//^O~ŕ\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DescTagActionExecute(TObject *Sender)
{
	DescTagAction->Checked = !DescTagAction->Checked;
	UpdateBranchList();
}

//---------------------------------------------------------------------------
//̃u`\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ShowBranchesActionExecute(TObject *Sender)
{
	ShowBranchesAction->Checked = !ShowBranchesAction->Checked;
	UpdateCommitList();
}
//---------------------------------------------------------------------------
//[gQƂ\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ShowRemoteActionExecute(TObject *Sender)
{
	ShowRemoteAction->Checked = !ShowRemoteAction->Checked;
	UpdateCommitList();
}
//---------------------------------------------------------------------------
//Author ̖O\
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ShowAuthorActionExecute(TObject *Sender)
{
	ShowAuthorAction->Checked = !ShowAuthorAction->Checked;
	UpdateCommitList();
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::GitPopupMenuPopup(TObject *Sender)
{
	reduction_MenuLine(((TPopupMenu*)Sender)->Items);
}

//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindCommitEditKeyDown(TObject *Sender, WORD &Key, TShiftState Shift)
{
	TListBox *lp = CommitListBox;
	int c_idx = lp->ItemIndex;

	UnicodeString KeyStr = get_KeyStr(Key, Shift);
	if		(contained_wd_i(KeysStr_CsrUp,   KeyStr)) { FindUpAction->Execute();	Key = 0; }
	else if (contained_wd_i(KeysStr_CsrDown, KeyStr)) { FindDownAction->Execute();	Key = 0; }
	else if (contained_wd_i(KeysStr_ToList,  KeyStr) || equal_ENTER(KeyStr)) lp->SetFocus();
	else return;

	if (c_idx!=lp->ItemIndex) lp->Invalidate();
	KeyHandled = true;
	Key = 0;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindCommitEditKeyPress(TObject *Sender, System::WideChar &Key)
{
	if (KeyHandled || Key==VK_RETURN || Key==VK_ESCAPE) {
		KeyHandled = false;
		Key = 0;
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindCommitEditChange(TObject *Sender)
{
	TListBox *lp = CommitListBox;
	UnicodeString kwd = FindCommitEdit->Text;
	if (!kwd.IsEmpty()) {
		int idx = -1;
		for (int i=0; i<lp->Count && idx==-1; i++)
			if (ContainsText(lp->Items->Strings[i], kwd)) idx = i;
		SetCommitListIndex(idx);

		if (CommitScrPanel->KeyWordChanged(kwd, lp->Count)) {
			for (int i=0; i<lp->Count; i++) {
				if (ContainsText(lp->Items->Strings[i], kwd)) CommitScrPanel->AddHitLine(i);
			}
			CommitScrPanel->Repaint();
		}
	}
	else {
		lp->ItemIndex = -1;
		CommitScrPanel->KeyWordChanged(EmptyStr, 0);
	}

	lp->Invalidate();
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindCommitEditEnter(TObject *Sender)
{
	set_EditColor(FindCommitEdit, true);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindCommitEditExit(TObject *Sender)
{
	CloseIME(Handle);
	InvColIfEmpty(FindCommitEdit);
}

//---------------------------------------------------------------------------
//R~bg
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindUpActionExecute(TObject *Sender)
{
	TListBox *lp = CommitListBox;
	UnicodeString kwd = FindCommitEdit->Text;
	int idx = -1;
	if (!kwd.IsEmpty() && lp->ItemIndex>0) {
		for (int i=lp->ItemIndex-1; i>=0 && idx==-1; i--)
			if (ContainsText(lp->Items->Strings[i], kwd)) idx = i;
	}
	SetCommitListIndex(idx);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindUpActionUpdate(TObject *Sender)
{
	TListBox *lp = CommitListBox;
	UnicodeString kwd = FindCommitEdit->Text;
	bool found = false;
	if (!kwd.IsEmpty() && lp->ItemIndex>0) {
		for (int i=lp->ItemIndex-1; i>=0 && !found; i--)
			found = ContainsText(lp->Items->Strings[i], kwd);
	}
	((TAction*)Sender)->Enabled = found;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindDownActionExecute(TObject *Sender)
{
	TListBox *lp = CommitListBox;
	UnicodeString kwd = FindCommitEdit->Text;
	int idx = -1;
	if (!kwd.IsEmpty() && lp->ItemIndex<lp->Count-1) {
		for (int i=lp->ItemIndex+1; i<lp->Count && idx==-1; i++)
			if (ContainsText(lp->Items->Strings[i], kwd)) idx = i;
	}
	SetCommitListIndex(idx);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindDownActionUpdate(TObject *Sender)
{
	TListBox *lp = CommitListBox;
	UnicodeString kwd = FindCommitEdit->Text;
	bool found = false;
	if (!kwd.IsEmpty() && lp->ItemIndex<lp->Count-1) {
		for (int i=lp->ItemIndex+1; i<lp->Count && !found; i++)
			found = ContainsText(lp->Items->Strings[i], kwd);
	}
	((TAction*)Sender)->Enabled = found;
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::FindBtnClick(TObject *Sender)
{
	FindCommitEdit->SetFocus();
}

//---------------------------------------------------------------------------
//XV
//---------------------------------------------------------------------------
void __fastcall TGitViewer::UpdateLogActionExecute(TObject *Sender)
{
	FilterName = EmptyStr;
	UpdateCommitList();
}

//---------------------------------------------------------------------------
//Diffc[N
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffToolActionExecute(TObject *Sender)
{
	git_rec *gp = GetCurCommitPtr();		if (!gp) return;
	UnicodeString fnam = GetDiffFileName();	if (fnam.IsEmpty()) return;

	UnicodeString fnam2;
	UnicodeString fnam1 = get_GitDiffFiles(fnam, fnam2);
	UnicodeString prm = "difftool -y";

	if (gp->is_stash) {
		prm.cat_sprintf(_T(" %s %s"), gp->stash.c_str(), fnam2.c_str());
	}
	else if (gp->is_work) {
		prm.cat_sprintf(_T(" %s"), fnam2.c_str());
	}
	else if (gp->is_index) {
		prm.cat_sprintf(_T(" --cached %s"), fnam2.c_str());
	}
	else {
		prm.cat_sprintf(_T(" %s:%s %s:%s"), get_tkn(ParentID, ' ').c_str(), fnam1.c_str(), CommitID.c_str(), fnam2.c_str());
	}

	if (!GitShellExe(prm, WorkDir)) msgbox_ERR(USTR_FaildExec);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::DiffActionUpdate(TObject *Sender)
{
	git_rec *gp = GetCurCommitPtr();
	((TAction *)Sender)->Enabled = !GetDiffFileName().IsEmpty()
			&& ((gp && (gp->is_stash || gp->is_work || gp->is_index)) || (!CommitID.IsEmpty() && !ParentID.IsEmpty()));
}

//---------------------------------------------------------------------------
//Git GREP J
//---------------------------------------------------------------------------
void __fastcall TGitViewer::GrepActionExecute(TObject *Sender)
{
	if (!GitGrepForm) GitGrepForm = new TGitGrepForm(this);	//ɓI쐬
	GitGrepForm->WorkDir  = WorkDir;
	GitGrepForm->CommitID = CommitID;
	GitGrepForm->ShowModal();
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::GrepActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction*)Sender;
	ap->Enabled = GetCurCommitPtr();
}

//---------------------------------------------------------------------------
//git-guiN
//---------------------------------------------------------------------------
void __fastcall TGitViewer::GuiActionExecute(TObject *Sender)
{
	UnicodeString prm = add_quot_if_spc(GitGuiExe);
	if (is_quot(prm)) prm.Insert("\"\" ", 1);
	if (!Execute_ex("cmd", "/c start " + prm, WorkDir, "H")) msgbox_ERR(USTR_FaildExec);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::GuiActionUpdate(TObject *Sender)
{
	((TAction*)Sender)->Enabled = !GitGuiExe.IsEmpty();
}

//---------------------------------------------------------------------------
//ConsoleN
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ConsoleActionExecute(TObject *Sender)
{
	UnicodeString prm = add_quot_if_spc(GitBashExe);
	if (is_quot(prm)) prm.Insert("\"\" ", 1);
	if (!Execute_ex("cmd", "/c start " + prm, WorkDir, "H")) msgbox_ERR(USTR_FaildExec);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ConsoleActionUpdate(TObject *Sender)
{
	((TAction*)Sender)->Enabled = !GitBashExe.IsEmpty();
}

//---------------------------------------------------------------------------
//eLXgGfB^ŊJ
//---------------------------------------------------------------------------
void __fastcall TGitViewer::EditFileActionExecute(TObject *Sender)
{
	git_rec *gp = GetCurCommitPtr();		if (!gp) return;
	UnicodeString fnam = GetDiffFileName();	if (fnam.IsEmpty()) return;
	fnam = get_GitDiffFile2(fnam);
	if (gp->is_work) {
		fnam = IncludeTrailingPathDelimiter(WorkDir) + slash_to_yen(fnam);
	}
	else {
		MsgHint->ShowMsgHint("t@Co...", this, col_bgHint);
		GitBusy  = true;
		fnam = save_GitRevAsTemp(gp->is_index? EmptyStr : CommitID, fnam, WorkDir);
		GitBusy  = false;
		MsgHint->ReleaseHandle();
	}
	if (!fnam.IsEmpty() && !open_by_TextEditor(fnam)) msgbox_ERR(GlobalErrMsg);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::EditFileActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	UnicodeString s = "eLXgGfB^ŊJ";
	s += (CommitID.IsEmpty()? "(&E)" : " - Temp (&E)");
	ap->Caption = s;
	ap->Enabled = !GetDiffTextName().IsEmpty();
}

//---------------------------------------------------------------------------
//GUIc[̑I
//---------------------------------------------------------------------------
void __fastcall TGitViewer::SelGuiItemClick(TObject *Sender)
{
	UserModule->PrepareOpenDlg(_T("GUIc[̑I"), F_FILTER_EXE2,
		ExtractFileName(GitGuiExe).c_str(), ExtractFileDir(GitGuiExe));
	UnicodeString fnam;
	if (UserModule->OpenDlgToStr(fnam)) GitGuiExe = fnam;
}
//---------------------------------------------------------------------------
//Console ̑I
//---------------------------------------------------------------------------
void __fastcall TGitViewer::SelConsoleItemClick(TObject *Sender)
{
	UserModule->PrepareOpenDlg(_T("Console ̑I"), F_FILTER_EXE2,
		ExtractFileName(GitBashExe).c_str(), ExtractFileDir(GitBashExe));
	UnicodeString fnam;
	if (UserModule->OpenDlgToStr(fnam)) GitBashExe = fnam;
}

//---------------------------------------------------------------------------
//ǉ
//---------------------------------------------------------------------------
void __fastcall TGitViewer::AddActionExecute(TObject *Sender)
{
	UnicodeString fnam = GetDiffFileName(true);
	if (!fnam.IsEmpty()) GitExeStr("add " + add_quot_if_spc(fnam));
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::AddActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	git_rec *gp = GetCurCommitPtr();
	ap->Enabled = (gp && gp->is_work && !GetDiffFileName(true).IsEmpty());
}

//---------------------------------------------------------------------------
//ׂĒǉ
//---------------------------------------------------------------------------
void __fastcall TGitViewer::AddAActionExecute(TObject *Sender)
{
	GitExeStr("add -A");
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::AddUActionExecute(TObject *Sender)
{
	GitExeStr("add -u");
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::AddAllActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	git_rec *gp = GetCurCommitPtr();
	ap->Visible = (gp && gp->is_work);
	ap->Enabled = ap->Visible;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ResetActionExecute(TObject *Sender)
{
	UnicodeString fnam = GetDiffFileName(true);
	if (!fnam.IsEmpty()) {
		//HEAD ̑݊mF
		UnicodeString prm = "rev-parse --verify --quiet HEAD";
		GitBusy = true;
		DWORD exit_code;
		bool ok = (GitShellExe(prm, WorkDir, (TStringList *)NULL, &exit_code) && exit_code==0);
		GitBusy = false;
		//
		prm = (ok? "reset HEAD " : "rm --cached ") + add_quot_if_spc(fnam);
		GitExeStr(prm);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ResetActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	git_rec *gp = GetCurCommitPtr();
	ap->Enabled = (gp && gp->is_index && !GetDiffFileName(true).IsEmpty());
}
//---------------------------------------------------------------------------
//ׂĎ
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ResetAllActionExecute(TObject *Sender)
{
	//HEAD ̑݊mF
	UnicodeString prm = "rev-parse --verify --quiet HEAD";
	GitBusy = true;
	DWORD exit_code;
	bool ok = (GitShellExe(prm, WorkDir, (TStringList *)NULL, &exit_code) && exit_code==0);
	GitBusy = false;
	GitExeStr(ok? "reset HEAD" : "rm --cached -r .");
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::ResetAllActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	git_rec *gp = GetCurCommitPtr();
	ap->Visible = (gp && gp->is_index);
	ap->Enabled = ap->Visible;
}

//---------------------------------------------------------------------------
//R~bg
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitActionExecute(TObject *Sender)
{
	if (!SetGitTagDlg) SetGitTagDlg = new TSetGitTagDlg(this);	//ɓI쐬
	SetGitTagDlg->IsCommit	 = true;
	SetGitTagDlg->EditMsgFile = IncludeTrailingPathDelimiter(WorkDir) + ".git\\COMMIT_EDITMSG";
	if (SetGitTagDlg->ShowModal() == mrOk) GitExeStr(SetGitTagDlg->GitParam);
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::CommitActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Enabled = Staged;
}
//---------------------------------------------------------------------------
//Ҕ
//---------------------------------------------------------------------------
void __fastcall TGitViewer::StashActionExecute(TObject *Sender)
{
	UnicodeString msg;
	if (input_query_ex(_T("Ҕ (stash)"), _T("bZ[W"), &msg)) {
		UnicodeString prm;
		prm.sprintf(_T("stash save \"%s\""), msg.c_str());
		GitExeStr(prm);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::StashActionUpdate(TObject *Sender)
{
	((TAction *)Sender)->Enabled = (StatusList->Count>0);
}

//---------------------------------------------------------------------------
//A/폜
//---------------------------------------------------------------------------
void __fastcall TGitViewer::StashPopActionExecute(TObject *Sender)
{
	if (!StashName.IsEmpty()) {
		UnicodeString msg;
		msg.sprintf(_T("[%s] 𕜋A/폜܂?"), StashName.c_str());
		if (msgbox_Sure(msg)) GitExeStr("stash pop " + StashName);
	}
}
//---------------------------------------------------------------------------
//A
//---------------------------------------------------------------------------
void __fastcall TGitViewer::StashApplyActionExecute(TObject *Sender)
{
	if (!StashName.IsEmpty()) {
		UnicodeString msg;
		msg.sprintf(_T("[%s] 𕜋A܂?"), StashName.c_str());
		if (msgbox_Sure(msg)) GitExeStr("stash apply " + StashName);
	}
}
//---------------------------------------------------------------------------
//폜
//---------------------------------------------------------------------------
void __fastcall TGitViewer::StashDropActionExecute(TObject *Sender)
{
	if (!StashName.IsEmpty()) {
		UnicodeString msg;
		msg.sprintf(_T("[%s] 폜܂?"), StashName.c_str());
		if (msgbox_Sure(msg)) GitExeStr("stash drop " + StashName);
	}
}
//---------------------------------------------------------------------------
void __fastcall TGitViewer::StashPopActionUpdate(TObject *Sender)
{
	TAction *ap = (TAction *)Sender;
	ap->Visible = !StashName.IsEmpty();
	ap->Enabled = ap->Visible;
}
//---------------------------------------------------------------------------
